//****************************************************************************
//                      Computing the complexity marker for a first-return map
//
//                                                       Rouen, March 20, 2024
//
// It is framed for 50,000 points in the first-return map saved in the 
// "response.dat" file (two columns). It returns the permutation entropy, the 
// structurality and the dynamical complexity.
//****************************************************************************
# include <cstdlib>
# include <iostream>
# include <fstream>
# include <iomanip>
# include <cmath>
# include <ctime>

  const int m = 6;                                    // state space dimension 

using namespace std;

double *rosdiade (int m,double x[]);

int main ( )
{
  ifstream data_12 ("response.dat");             // First-return map data file
  ifstream data_14 ("response.dat");
  int i, j, k, l, n, saveq, bfive, bten, px, py;
  int *seq, *q, *trans, *trans6, *trans10;
  int Ns = 0;                                  // Number of possible sequences
  int Npix;                                        // number of visited pixels
  const int Npt = 50000;                              // number of data points
  const int Nq = 6;                                     // Length of sequences
  const int Nmax = int(pow(Nq,Nq));            // Number of possible sequences
  const int Npixel = max(int(sqrt(float(Npt/40))),20);
  double Xmin = +1.0E+5;
  double Xmax = -1.0E+5;
  double dX = (Xmax-Xmin)/double(Npixel);
  double sx, sy, Sp, Comp, Struc, Smax, Pmin, Pmax; 
  double *x, *ser, *P, xp, yp; 
  bool pass[Npixel][Npixel];

  seq = new int[Nq];
  ser = new double[Nq];
  q = new int[Nq];
  trans = new int[Nmax];
  trans6 = new int[Nmax];
  trans10 = new int[Nmax];

  cout << "\n";
  cout << " Computing the complexity marker from a first-return map \n";
  cout << Npixel << " X " << Npixel << " pixels \n \n";

// ------------------ Initializing the possible sequences ---------------------
  for (q[0] = 0; q[0] < Nq; q[0]++)
  {
    for (q[1] = 0; q[1] < Nq; q[1]++)
    {
      if (q[1] != q[0])
      {
        for (q[2] = 0; q[2] < Nq; q[2]++)
        {
          if (q[2] != q[1] && q[2] != q[0])
          {
            for (q[3] = 0; q[3] < Nq; q[3]++)
            {
              if (q[3] != q[2] && q[3] != q[1] && q[3] != q[0])
              {
                for (q[4] = 0; q[4] < Nq; q[4]++)
                {
                  if (q[4] != q[3] && q[4] != q[2] && q[4] != q[1]
                                   && q[4] != q[0])
                  {
                    for (q[5] = 0; q[5] < Nq; q[5]++)
                    {
                      if (q[5] != q[4] && q[5] != q[3] && q[5] != q[2]
                                       && q[5] != q[1] && q[5] != q[0])
                      {
                        bfive = 0;
                        bten = 0;
                        for (l = 0; l < Nq; l++)
                        {
                          bfive += q[Nq-1-l]*pow(Nq,l);
                          bten += q[Nq-1-l]*pow(10,l);
                        }
                        trans[bfive] = Ns;
                        trans6[Ns] = bfive;
                        trans10[Ns] = bten;
                        Ns++;
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  P = new double[Ns];
  cout << "There are " << Ns << " actually realized sequences \n \n";
//--------- Initializiation of the possible sequences completed --------------
  Ns = 1;
  for (j = 2; j <= Nq ; ++j ) Ns *= j;
  cout << "Ns = " << Ns << "\n";
  Smax = log(double(Ns));
  cout << "Sp max = " << Smax << "\n";
  cout << "Normalization completed! \n \n";
//------------------- Normalization completed --------------------------------
  for (n = 0; n < Npt; n++)
  {
    data_12 >> xp, yp;
    Xmin = min(Xmin,xp);
    Xmax = max(Xmax,xp);
  }
  data_12.close ( );
  Xmin -= 0.05*fabs(Xmin);
  Xmax += 0.05*fabs(Xmax);
  cout << Xmin << " < x < " << Xmax << "\n";
  dX = (Xmax-Xmin)/double(Npixel);
  cout << " dX = " << dX << "\n";
    for (px = 0; px < Npixel; px++)
    {
      for (py = 0; py < Npixel; py++)
      {
        pass[px][py] = true;
      }
    }
    Npix = 0;
    for (j = 0; j < Ns; j++)
    {
      P[j] = 0.0;
    } 
    j = 0;
    n = 0;
    do 
    {
      n++;
      for (k = 0; k < Nq-1; k++)
      {
        ser[k] = ser[k+1];
      }
      data_14 >> ser[Nq-1], yp;
    } while (n < Nq);
    do 
    {
      n++;
      for (k = 0; k < Nq-1; k++)
      {
        ser[k] = ser[k+1];
      }
      data_14 >> ser[Nq-1],yp;
      px = int((ser[Nq-2]-Xmin)/dX);
      py = int((ser[Nq-1]-Xmin)/dX);
      if (pass[px][py])
      {
        pass[px][py] = false;
        Npix++;
      }
      for (i = 0; i < Nq; i++)
      {
        seq[i] = i;
      }
      for (k = 1; k < Nq; k++)
      {
        i = k;
        do
        {
          if (ser[seq[i]] < ser[seq[i-1]])
          {
            saveq = seq[i-1];
            seq[i-1] = seq[i];
            seq[i] = saveq;
          }
          else
          {
            i = 1;
          }
          i--;
        } while (i > 0);
      }
      bfive = 0;
      for (l = 0; l < Nq; l++)
      {
        bfive += seq[Nq-1-l]*pow(Nq,l);
      }
      P[trans[bfive]] += 1.0; 
    } while (n < Npt);
    i = 0;
    Sp = 0.0;
    for (j = 0; j < Ns; j++)
    {
      if (P[j] != 0.0)
      {
        i++;
        bfive = trans6[j];
        bten = trans10[j];
        P[j] /= double(Npt);
        Sp += -P[j]*log(P[j]);
      }
    }
    Sp /= Smax;
    Struc = double(Npix)/double(Npixel*Npixel);
    Comp = Struc+Sp;
    cout << " Shannon entropy Sp = " << Sp << "\n";
    cout << " Structurality D = " << Struc << " Npix = " << Npix << "\n";
    cout << " Complexity C_D = " << Comp << "\n"; 
  data_14.close ( );
  return 0;
}
// ***************************************************************************
